import codecs
from os import path

here = path.dirname(path.abspath(__file__))


def get_resource(name):
    resource_dir = path.join(here, "resources")
    pth = path.join(resource_dir, name)
    with codecs.open(pth, "r", encoding="utf-8") as f:
        return f.read()



import typing
from abc import abstractmethod

import attr

from applitools.common import CoordinatesType, FloatingMatchSettings
from applitools.common.accessibility import AccessibilityRegionType
from applitools.common.geometry import AccessibilityRegion, Point, Region
from applitools.common.utils import ABC
from applitools.core import GetFloatingRegion, GetRegion
from applitools.core.fluent.region import GetAccessibilityRegion
from applitools.selenium.capture import EyesWebDriverScreenshot

if typing.TYPE_CHECKING:
    from typing import List, Optional

    from applitools.common import FloatingBounds
    from applitools.common.utils.custom_types import (
        AnyWebDriver,
        AnyWebElement,
        CodedRegionPadding,
    )
    from applitools.selenium.selenium_eyes import SeleniumEyes

__all__ = (
    "RegionBySelector",
    "RegionByElement",
    "FloatingRegionBySelector",
    "FloatingRegionByElement",
    "AccessibilityRegionBySelector",
    "AccessibilityRegionByElement",
)


def _region_from_element(element, screenshot, padding=None):
    # type: (AnyWebElement,EyesWebDriverScreenshot,Optional[CodedRegionPadding])->Region
    location = element.location
    if screenshot:
        # Element's coordinates are context relative, so we need to convert them first.
        adjusted_location = screenshot.location_in_screenshot(
            Point.from_(location), CoordinatesType.CONTEXT_RELATIVE
        )
    else:
        adjusted_location = Point.from_(location)
    region = Region.from_(adjusted_location, element.size)

    if padding:
        region = region.padding(padding)
    return region


class GetSeleniumRegion(GetRegion, ABC):
    def get_regions(self, eyes, screenshot):
        # type: (SeleniumEyes, EyesWebDriverScreenshot) -> List[Region]
        elements = self._fetch_elements(eyes.driver)
        return [_region_from_element(el, screenshot, self._padding) for el in elements]

    def get_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return self._fetch_elements(driver)

    @abstractmethod
    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        pass


@attr.s
class RegionByElement(GetSeleniumRegion):
    _element = attr.ib()  # type: AnyWebElement
    _padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]

    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return [self._element]


@attr.s
class RegionBySelector(GetSeleniumRegion):
    """
    :param by: The "by" part of a selenium selector for an element which
               represents the ignore region
    :param value: The "value" part of a selenium selector for
                  an element which represents the ignore region.
    """

    _by = attr.ib()
    _value = attr.ib()
    _padding = attr.ib(default=None)  # type: Optional[CodedRegionPadding]

    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return driver.find_elements(self._by, self._value)


class GetSeleniumFloatingRegion(GetFloatingRegion, ABC):
    def get_regions(self, eyes, screenshot):
        # type: (SeleniumEyes, EyesWebDriverScreenshot) ->  List[FloatingMatchSettings]
        elements = self._fetch_elements(eyes.driver)
        regions = (_region_from_element(el, screenshot) for el in elements)
        return [
            FloatingMatchSettings(region, self.floating_bounds) for region in regions
        ]

    def get_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return self._fetch_elements(driver)

    @abstractmethod
    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        pass


@attr.s
class FloatingRegionByElement(GetSeleniumFloatingRegion):
    """
    :ivar element: The element which represents the inner region (the floating part).
    :ivar bounds: The outer rectangle bounding the inner region.
    """

    _element = attr.ib()  # type: AnyWebElement
    _bounds = attr.ib()  # type: FloatingBounds

    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return [self._element]

    @property
    def floating_bounds(self):
        # type: () -> FloatingBounds
        return self._bounds


@attr.s
class FloatingRegionBySelector(GetSeleniumFloatingRegion):
    """
    :ivar by: The selenium By
    :ivar value: The css selector for an element which represents the inner region.
    :ivar bounds: The outer rectangle bounding the inner region.
    """

    _by = attr.ib()  # type: str
    _value = attr.ib()  # type: str
    _bounds = attr.ib()  # type: FloatingBounds

    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return driver.find_elements(self._by, self._value)

    @property
    def floating_bounds(self):
        # type: () -> FloatingBounds
        return self._bounds


class GetSeleniumAccessibilityRegion(GetAccessibilityRegion, ABC):
    def get_regions(self, eyes, screenshot):
        # type:(SeleniumEyes,EyesWebDriverScreenshot)->List[AccessibilityRegion]
        elements = self._fetch_elements(eyes.driver)
        regions = (_region_from_element(el, screenshot) for el in elements)
        return [
            AccessibilityRegion.from_(region, self.accessibility_type)
            for region in regions
        ]

    def get_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return self._fetch_elements(driver)

    @abstractmethod
    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        pass


@attr.s
class AccessibilityRegionBySelector(GetSeleniumAccessibilityRegion):
    _by = attr.ib()  # type: str
    _value = attr.ib()  # type: str
    _type = attr.ib()  # type: AccessibilityRegionType

    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return driver.find_elements(self._by, self._value)

    @property
    def accessibility_type(self):
        # type: () -> AccessibilityRegionType
        return self._type


@attr.s
class AccessibilityRegionByElement(GetSeleniumAccessibilityRegion):
    _element = attr.ib()  # type: AnyWebElement
    _type = attr.ib()  # type: AccessibilityRegionType

    def _fetch_elements(self, driver):
        # type: (AnyWebDriver) -> List[AnyWebElement]
        return [self._element]

    @property
    def accessibility_type(self):
        # type: () -> AccessibilityRegionType
        return self._type



import sys
import threading
import typing
from concurrent.futures import Future, ThreadPoolExecutor
from functools import partial

from applitools.common import VGResource, logger
from applitools.common.utils.converters import str2bool
from applitools.common.utils.general_utils import get_env_with_prefix
from applitools.core import ServerConnector

if typing.TYPE_CHECKING:
    from typing import List


class ResourceCache(typing.Mapping[typing.Text, VGResource]):
    def __init__(self):
        self.cache_map = {}
        kwargs = {}
        if sys.version_info[:2] >= (3, 6):
            kwargs["thread_name_prefix"] = "ResourceCache-Executor"
        self.executor = ThreadPoolExecutor(**kwargs)
        self.lock = threading.RLock()

    def __str__(self):
        with self.lock:
            return str(self.cache_map)[:25]

    def _process_future(self, url, val):
        if isinstance(val, Future):
            try:
                val = val.result()
            except Exception:
                logger.exception(
                    "We got an exception during processing URL: {}".format(url)
                )
                val = None
            finally:
                self.cache_map[url] = val
        return val

    def __getitem__(self, item):
        with self.lock:
            val = self.cache_map[item]
            return self._process_future(item, val)

    def __setitem__(self, key, value):
        with self.lock:
            self.cache_map[key] = value

    def __len__(self):
        with self.lock:
            return len(self.cache_map)

    def __del__(self):
        with self.lock:
            self.executor.shutdown()

    def __iter__(self):
        with self.lock:
            return iter(list(self.cache_map))

    def fetch_and_store(self, url, fetch_function, force=False):
        # type: (typing.Text, typing.Callable[[typing.Text], VGResource], bool) -> bool
        """
        Schedules fetch of the url using fetch_function.

        :param url: url to fetch resource from
        :param fetch_function: function called to fetch the resource in a worker thread
        :param force: option to force fetch even if the resource is
                      already fetched or fetching

        :return: True if the fetch was just scheduled or already in process
                 False if it's already fetched and in the cache
        """
        with self.lock:
            if force or url not in self.cache_map:
                self.cache_map[url] = self.executor.submit(fetch_function, url)
                return True
            else:
                return isinstance(self.cache_map[url], Future)

    def process_all(self):
        with self.lock:
            return [self[r_url] for r_url in self]


class PutCache(object):
    def __init__(self):
        self._force_put = str2bool(
            get_env_with_prefix("APPLITOOLS_UFG_FORCE_PUT_RESOURCES")
        )
        kwargs = {}
        if sys.version_info[:2] >= (3, 6):
            kwargs["thread_name_prefix"] = self.__class__.__name__
        self._executor = ThreadPoolExecutor(**kwargs)
        self._lock = threading.Lock()
        self._sent_hashes = set()

    def put(
        self,
        all_resources,  # type: List[VGResource]
        server_connector,  # type: ServerConnector
    ):
        # type: (...) -> None
        logger.debug("PutCache.put({} call".format(all_resources))
        with self._lock:
            not_puted = [r for r in all_resources if r.hash not in self._sent_hashes]
            if not self._force_put:
                check_result = server_connector.check_resource_status(not_puted)
                resources_to_upload = []
                for resource, exists in zip(not_puted, check_result):
                    if exists:
                        self._sent_hashes.add(resource.hash)
                        resource.clear()
                        continue
                    if resource.content is None:
                        logger.warning(
                            "This resource was requested by server but is cleared: "
                            "{} {}".format(resource.hash, resource.url)
                        )
                    else:
                        resources_to_upload.append(resource)
            else:
                resources_to_upload = not_puted
            results_iterable = self._executor.map(
                server_connector.render_put_resource, resources_to_upload
            )
            self._sent_hashes.update(list(results_iterable))
            for resource in resources_to_upload:
                resource.clear()

    def shutdown(self):
        # type: () -> None
        with self._lock:
            self._executor.shutdown()

    def __del__(self):
        # type: () -> None
        self.shutdown()



from .eyes_connector import EyesConnector
from .resource_cache import PutCache, ResourceCache
from .resource_collection_task import ResourceCollectionTask
from .running_test import RunningTest
from .vg_task import VGTask
from .visual_grid_eyes import VisualGridEyes
from .visual_grid_runner import RunnerOptions, VisualGridRunner

__all__ = (
    "VGTask",
    "PutCache",
    "ResourceCache",
    "VisualGridEyes",
    "EyesConnector",
    "VisualGridRunner",
    "RunnerOptions",
    "RunningTest",
    "ResourceCollectionTask",
)



import typing
from copy import deepcopy
from threading import Condition, Thread
from time import time

import attr

from applitools.common import EyesError, RenderStatus, logger
from applitools.common.utils import datetime_utils

if typing.TYPE_CHECKING:
    from typing import Callable, List, Text

    from applitools.common import RenderRequest, RenderStatusResults
    from applitools.core import ServerConnector


MAX_STATUS_CHECK_SECONDS = 60 * 60


class RenderingService(object):
    def __init__(self):
        self._server_connector = None  # type: ServerConnector
        self._shutdown = False
        self._render_tasks = []  # type: List[_Render]
        self._have_render_tasks = Condition()
        self._status_tasks = []  # type: List[_Status]
        self._have_status_tasks = Condition()
        self._render_request_thread = Thread(target=self._run_render_requests)
        self._render_request_thread.setName("RenderingBatching")
        self._render_request_thread.daemon = True
        self._render_status_thread = Thread(target=self._run_render_status_requests)
        self._render_status_thread.setName("RenderingStatusBatching")
        self._render_status_thread.daemon = True
        self._render_request_thread.start()
        self._render_status_thread.start()

    def maybe_set_server_connector(self, server_connector):
        # type: (ServerConnector) -> None
        if not self._server_connector:
            self._server_connector = server_connector

    def render(self, render_request, on_success, on_error):
        # type: (RenderRequest, Callable, Callable) -> None
        with self._have_render_tasks:
            self._render_tasks.append(_Render(render_request, on_success, on_error))
            self._have_render_tasks.notify()

    def shutdown(self):
        self._shutdown = True
        with self._have_render_tasks:
            self._have_render_tasks.notify()
        with self._have_status_tasks:
            self._have_status_tasks.notify()
        self._render_request_thread.join()
        self._render_status_thread.join()

    def _run_render_requests(self):
        while True:
            try:
                with self._have_render_tasks:
                    while not (self._shutdown or self._render_tasks):
                        self._have_render_tasks.wait()
                    if self._shutdown:
                        break
                    render_tasks, self._render_tasks = self._render_tasks, []
                requests = [r.request for r in render_tasks]
                logger.debug("render call with requests: {}".format(requests))
                try:
                    responses = self._server_connector.render(*requests)
                except Exception as e:
                    for task in render_tasks:
                        task.on_error(e)
                    continue
                statuses = []
                for task, running_render in zip(render_tasks, responses):
                    if running_render.render_status == RenderStatus.RENDERING:
                        statuses.append(
                            _Status(
                                running_render.render_id, task.on_success, task.on_error
                            )
                        )
                    else:
                        task.on_error(
                            EyesError("Unexpected render response", running_render)
                        )
                with self._have_status_tasks:
                    self._status_tasks.extend(statuses)
                    self._have_status_tasks.notify()
            except Exception:
                logger.exception("Render request batching thread failure")
                datetime_utils.sleep(5000, "Render request error delay")

    def _run_render_status_requests(self):
        status_tasks = []
        while True:
            try:
                with self._have_status_tasks:
                    while not (self._shutdown or status_tasks or self._status_tasks):
                        self._have_status_tasks.wait()
                    if self._shutdown:
                        break
                    new_tasks, self._status_tasks = self._status_tasks, []
                status_tasks.extend(new_tasks)
                status_tasks = self._check_render_status_batch(status_tasks)
                if status_tasks:
                    datetime_utils.sleep(1000, "Render status polling delay")
            except Exception:
                logger.exception("Rendering status batching thread failure")
                datetime_utils.sleep(5000, "Render status error delay")

    def _check_render_status_batch(self, status_tasks):
        ids = [t.render_id for t in status_tasks]
        logger.debug("render_status_by_id call with ids: {}".format(ids))
        try:
            statuses = self._server_connector.render_status_by_id(*ids)
            logger.debug(
                "Received render status response", statuses_count=len(statuses)
            )
            status_by_render_id = {s.render_id: s for s in statuses if s}
        except Exception as e:
            for task in status_tasks:
                task.on_error(e)
            return []
        rendering = []
        current_time = time()
        for task in status_tasks:
            status = status_by_render_id.get(task.render_id)
            if status is None:
                logger.error(
                    "Missing render id in RenderStatus response",
                    render_id=task.render_id,
                )
                task.on_error(EyesError("Unknown rendering error"))
            elif status.status == RenderStatus.RENDERING:
                if current_time > task.timeout_time:
                    task.on_error(EyesError("Render time out"))
                else:
                    rendering.append(task)
            else:
                task.on_success(status)
        return rendering


@attr.s
class _Render(object):
    request = attr.ib()  # type: RenderRequest
    on_success = attr.ib()  # type: Callable[[RenderStatusResults], None]
    on_error = attr.ib()  # type: Callable[[Exception], None]


@attr.s
class _Status(object):
    render_id = attr.ib()  # type: Text
    on_success = attr.ib()  # type: Callable[[RenderStatusResults], None]
    on_error = attr.ib()  # type: Callable[[Exception], None]
    timeout_time = attr.ib(
        factory=lambda: time() + MAX_STATUS_CHECK_SECONDS
    )  # type: int



from typing import TYPE_CHECKING, Generator

import tinycss2
from lxml import etree

from applitools.common import logger

if TYPE_CHECKING:
    from typing import List, Text


def _url_from_tags(tags, url_tag_types):
    for tag in tags:
        if tag.type in url_tag_types:
            try:
                yield tag.value
            except Exception as e:
                logger.exception(e)


def get_urls_from_css_resource(bytes_text):
    # type: (bytes) -> List[Text]
    def is_import_node(n):
        return n.prelude and n.type == "at-rule" and n.lower_at_keyword == "import"

    try:
        rules, encoding = tinycss2.parse_stylesheet_bytes(
            css_bytes=bytes_text, skip_comments=True, skip_whitespace=True
        )
    except Exception:
        logger.error("Failed to read CSS string")
        return []
    urls = []
    for rule in rules:
        if is_import_node(rule):
            extracted = _url_from_tags(rule.prelude, ("url", "string"))
        elif rule.content:
            extracted = _url_from_tags(rule.content, ("url",))
        else:
            continue
        urls.extend(list(extracted))
    return urls


def get_urls_from_svg_resource(content):
    # type: (bytes) -> List[Text]
    xmlparser = etree.XMLParser(recover=True, ns_clean=True, remove_comments=True)
    xml = etree.HTML(content, xmlparser)
    root = xml.getroottree()
    nodes = root.xpath(".//*[@href]") + root.xpath(
        ".//*[@xlink:href]", namespaces={"xlink": "http://www.w3.org/1999/xlink"}
    )
    urls_from_svg = []
    for node in nodes:
        # node.attrib could contains href w key with namespace
        key = [key for key in node.attrib.keys() if key.endswith("href")][0]
        url = node.attrib[key]
        if url.startswith("data:") and url.startswith("javascript:"):
            continue
        urls_from_svg.append(url)
    return urls_from_svg


def collect_urls_from_(content_type, content):
    # type: (Text, bytes) -> Generator[Text]
    urls_from_css, urls_from_svg = [], []
    if content_type.startswith("text/css"):
        urls_from_css = get_urls_from_css_resource(content)
    if content_type.startswith("image/svg"):
        urls_from_svg = get_urls_from_svg_resource(content)
    for discovered_url in urls_from_css + urls_from_svg:
        if discovered_url.startswith("data:") or discovered_url.startswith("#"):
            # resource already in blob or not relevant
            continue
        yield discovered_url



from appium.webdriver import WebElement as AppiumWebElement
from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement
from selenium.webdriver.support.event_firing_webdriver import EventFiringWebElement

from applitools.common.validators import *  # noqa
from applitools.selenium.webelement import EyesWebElement


def is_webelement(elm):
    def check(elem):
        return isinstance(
            elm,
            (
                EyesWebElement,
                SeleniumWebElement,
                AppiumWebElement,
                EventFiringWebElement,
            ),
        )

    return check(elm) or check(getattr(elm, "_element", None))



from time import time
from typing import TYPE_CHECKING, Optional, Union

from applitools.common import (
    DiffsFoundError,
    EyesError,
    NewTestError,
    TestFailedError,
    TestResultContainer,
    TestResults,
    logger,
)
from applitools.common.utils import datetime_utils, iteritems

if TYPE_CHECKING:
    from typing import Callable, Dict, List

    from applitools.selenium.visual_grid import RunningTest


def wait_till_tests_completed(test_provider, timeout):
    # type: (Union[Callable, List], Optional[int]) -> None
    def get_tests(provider):
        if isinstance(test_provider, list):
            return test_provider
        return test_provider()

    deadline = time() + timeout if timeout else None
    iterations = 0
    while deadline is None or time() < deadline:
        states = list(set(t.state for t in get_tests(test_provider)))
        if not states:
            # probably some exception is happened during execution
            break
        if len(states) == 1 and states[0] == "completed":
            break
        datetime_utils.sleep(1500, msg="Waiting for state completed!")
        iterations += 1
        if iterations % 200 == 0:
            logger.debug(
                "Unfinished tests state report",
                unfinished_tests=tests_state_report(
                    t for t in get_tests(test_provider) if t.state != "completed"
                ),
            )
    else:
        logger.warning(
            "Tests completion timeout exceeded",
            timeout=timeout,
            unfinished_tests=tests_state_report(
                t for t in get_tests(test_provider) if t.state != "completed"
            ),
        )
        raise EyesError("Tests didn't finish in {} seconds".format(timeout))


def tests_state_report(tests):
    state_report = []
    for test in tests:
        state_report.append(
            {
                "app_name": test.configuration.app_name,
                "test_name": test.configuration.test_name,
                "browser_info": test.browser_info,
                "uuid": test.test_uuid,
                "state": test.state,
                "active_check": test.task_lock,
            }
        )
    return state_report


def collect_test_results(tests, should_raise_exception):
    # type: (Dict[RunningTest, TestResults], bool) -> List[TestResultContainer]
    all_results = []
    for test, test_result in iteritems(tests):
        if test.pending_exceptions:
            logger.error(
                "During test execution above exception raised. \n {:s}".join(
                    str(e) for e in test.pending_exceptions
                )
            )
        if test.has_checks:
            exception = None
        else:
            exception = TestFailedError("Test has no checks")
        if test_result:
            scenario_id_or_name = test_result.name
            app_id_or_name = test_result.app_name
            if test_result.is_unresolved and not test_result.is_new:
                exception = DiffsFoundError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            if test_result.is_new:
                exception = NewTestError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            if test_result.is_failed:
                exception = TestFailedError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
        else:
            exception = TestFailedError("Test haven't finished correctly")
        all_results.append(
            TestResultContainer(test_result, test.browser_info, exception)
        )
        if exception and should_raise_exception:
            raise exception
    return all_results
